@@ -1,15 +1,15 @@
+import warnings
 from numbers import Number
-from typing import Callable, List, Tuple
+from typing import Callable, List, Tuple, Optional
 
 from phi import geom
 from phi import math
-from phi.geom import Box, Geometry, Sphere, Cuboid
-from phi.math import Tensor, spatial, instance, tensor, masked_fill, channel, Shape, batch, unstack, wrap, vec
+from phi.geom import Box, Geometry
+from phi.math import Tensor, spatial, instance, tensor, channel, Shape, unstack, solve_linear, jit_compile_linear, shape, Solve, extrapolation, jit_compile
 from ._field import Field, SampledField, SampledFieldType, as_extrapolation
 from ._grid import CenteredGrid, Grid, StaggeredGrid, GridType
 from ._point_cloud import PointCloud
-from .numerical import Scheme
-from ..math.extrapolation import Extrapolation
+from ..math.extrapolation import Extrapolation, SYMMETRIC, REFLECT, ANTIREFLECT, ANTISYMMETRIC, combine_by_direction
 
 
 def bake_extrapolation(grid: GridType) -> GridType:
@@ -38,18 +38,78 @@ def bake_extrapolation(grid: GridType) -> GridType:
         raise ValueError(f"Not a valid grid: {grid}")
 
 
-def laplace(field: GridType, axes=spatial, weights: Tensor or Field = None) -> GridType:
-    """ Finite-difference laplace operator for Grids. See `phi.math.laplace()`. """
+def laplace(field: GridType,
+            axes=spatial,
+            order=2,
+            implicit: math.Solve = None,
+            weights: Tensor or Field = None) -> GridType:
+    """
+    Spatial Laplace operator for scalar grid.
+    If a vector grid is passed, it is assumed to be centered and the laplace is computed component-wise.
+
+    Args:
+        field: n-dimensional `CenteredGrid`
+        axes: The second derivative along these dimensions is summed over
+        weights: (Optional) Multiply the axis terms by these factors before summation.
+            Must be a `phi.math.Tensor` or `phi.field.Field` with a single channel dimension that lists all laplace axes by name.
+        order: Spatial order of accuracy.
+            Higher orders entail larger stencils and more computation time but result in more accurate results assuming a large enough resolution.
+            Supported: 2 explicit, 4 explicit, 6 implicit.
+        implicit: When a `Solve` object is passed, performs an implicit operation with the specified solver and tolerances.
+            Otherwise, an explicit stencil is used.
+
+    Returns:
+        laplacian field as `CenteredGrid`
+    """
+    if implicit:
+        warnings.warn("Implicit operators currently do not support sparse matrix generation and may be slow.", RuntimeWarning, stacklevel=2)
     if isinstance(weights, Field):
         weights = weights.at(field).values
-    result = field._op1(lambda tensor: math.laplace(tensor, dx=field.dx, padding=field.extrapolation, dims=axes, weights=weights))
+    axes_names = field.shape.only(axes).names
+    extrap_map = {}
+    if not implicit:
+        if order == 2:
+                values, needed_shifts = [1, -2, 1], (-1, 0, 1)
+
+        elif order == 4:
+                values, needed_shifts = [-1/12, 4/3, -5/2, 4/3, -1/12], (-2, -1, 0, 1, 2)
+    else:
+        extrap_map_rhs = {}
+        if order == 6:
+            values, needed_shifts = [3/44, 12/11, -51/22, 12/11, 3/44], (-2, -1, 0, 1, 2)
+            extrap_map['symmetric'] = combine_by_direction(REFLECT, SYMMETRIC)
+            values_rhs, needed_shifts_rhs = [2/11, 1, 2/11], (-1, 0, 1)
+            extrap_map_rhs['symmetric'] = combine_by_direction(REFLECT, SYMMETRIC)
+    base_widths = (abs(min(needed_shifts)), max(needed_shifts))
+    field.with_extrapolation(extrapolation.map(_ex_map_f(extrap_map), field.extrapolation))
+    padded_components = [pad(field, {dim: base_widths}) for dim in axes_names]
+    shifted_components = [shift(padded_component, needed_shifts, None, pad=False, dims=dim) for padded_component, dim in zip(padded_components, axes_names)]
+    result_components = [sum([value * shift_ for value, shift_ in zip(values, shifted_component)]) / field.dx.vector[dim]**2 for shifted_component, dim in zip(shifted_components, axes_names)]
+    if implicit:
+        result_components = stack(result_components, channel('laplacian'))
+        result_components.with_values(result_components.values._cache())
+        result_components = result_components.with_extrapolation(extrapolation.map(_ex_map_f(extrap_map_rhs), field.extrapolation))
+        implicit.x0 = result_components
+        result_components = solve_linear(_lhs_for_implicit_scheme, result_components, solve=implicit, values_rhs=values_rhs, needed_shifts_rhs=needed_shifts_rhs, stack_dim=channel('laplacian'))
+        result_components = unstack(result_components, 'laplacian')
+        extrap_map = extrap_map_rhs
+    result_components = [component.with_bounds(field.bounds) for component in result_components]
+    if weights is not None:
+        assert channel(weights).rank == 1 and channel(weights).item_names is not None, f"weights must have one channel dimension listing the laplace dims but got {shape(weights)}"
+        assert set(channel(weights).item_names[0]) >= set(axes_names), f"the channel dim of weights must contain all laplace dims {axes_names} but only has {channel(weights).item_names}"
+        result_components = [c * weights[ax] for c, ax in zip(result_components, axes_names)]
+    result = sum(result_components)
+    result = result.with_extrapolation(extrapolation.map(_ex_map_f(extrap_map), field.extrapolation))
     return result
 
 
-def spatial_gradient(field: Field,
-                     extrapolation: Extrapolation = None,
+def spatial_gradient(field: CenteredGrid,
+                     gradient_extrapolation: Extrapolation = None,
                      type: type = CenteredGrid,
-                     stack_dim: Shape = channel('vector')):
+                     dims: math.DimFilter = spatial,
+                     stack_dim: Shape = channel('vector'),
+                     order=2,
+                     implicit: Solve = None):
     """
     Finite difference spatial_gradient.
 
@@ -60,29 +120,123 @@ def spatial_gradient(field: Field,
 
     Args:
         field: centered grid of any number of dimensions (scalar field, vector field, tensor field)
+        gradient_extrapolation: Extrapolation of the output
         type: either `CenteredGrid` or `StaggeredGrid`
+        dims: Along which dimensions to compute the spatial gradient. Only supported when `type==CenteredGrid`.
         stack_dim: Dimension to be added. This dimension lists the spatial_gradient w.r.t. the spatial dimensions.
             The `field` must not have a dimension of the same name.
+        order: Spatial order of accuracy.
+            Higher orders entail larger stencils and more computation time but result in more accurate results assuming a large enough resolution.
+            Supported: 2 explicit, 4 explicit, 6 implicit.
+        implicit: When a `Solve` object is passed, performs an implicit operation with the specified solver and tolerances.
+            Otherwise, an explicit stencil is used.
 
     Returns:
         spatial_gradient field of type `type`.
-
     """
-    assert isinstance(field, Grid)
-    if extrapolation is None:
-        extrapolation = field.extrapolation.spatial_gradient()
+    if implicit:
+        warnings.warn("Implicit operators currently do not support sparse matrix generation and may be slow.", RuntimeWarning, stacklevel=2)
+    if gradient_extrapolation is None:
+        gradient_extrapolation = field.extrapolation.spatial_gradient()
+    extrap_map = {}
+    if not implicit:
+        if order == 2:
+            if type == CenteredGrid:
+                values, needed_shifts = [-1/2, 1/2], (-1, 1)
+            else:
+                values, needed_shifts = [-1, 1], (0, 1)
+        elif order == 4:
+            if type == CenteredGrid:
+                values, needed_shifts = [1/12, -2/3, 2/3, -1/12], (-2, -1, 1, 2)
+            else:
+                values, needed_shifts = [1/24, -27/24, 27/24, -1/24], (-1, 0, 1, 2)
+        else:
+            raise NotImplementedError(f"explicit {order}th-order not supported")
+    else:
+        extrap_map_rhs = {}
+        if order == 6:
+            if type == CenteredGrid:
+                values, needed_shifts = [-1/36, -14/18, 14/18, 1/36], (-2, -1, 1, 2)
+                values_rhs, needed_shifts_rhs = [1/3, 1, 1/3], (-1, 0, 1)
+            else:
+                values, needed_shifts = [-17/186, -63/62, 63/62, 17/186], (-1, 0, 1, 2)
+                extrap_map['symmetric'] = combine_by_direction(REFLECT, SYMMETRIC)
+                values_rhs, needed_shifts_rhs = [9/62, 1, 9/62], (-1, 0, 1)
+                extrap_map_rhs['symmetric'] = combine_by_direction(ANTIREFLECT, ANTISYMMETRIC)
+        else:
+            raise NotImplementedError(f"implicit {order}th-order not supported")
+    base_widths = (abs(min(needed_shifts)), max(needed_shifts))
+    field.with_extrapolation(extrapolation.map(_ex_map_f(extrap_map), field.extrapolation))  # ToDo does this line do anything?
+    if implicit:
+        gradient_extrapolation = extrapolation.map(_ex_map_f(extrap_map_rhs), gradient_extrapolation)
+    spatial_dims = field.shape.only(dims).names
+    stack_dim = stack_dim._with_item_names((spatial_dims,))
     if type == CenteredGrid:
-        pad = 1 if extrapolation == math.extrapolation.NONE else 0
-        values = math.spatial_gradient(field.values, field.dx.vector.as_channel(name=stack_dim.name), difference='central', padding=field.extrapolation, stack_dim=stack_dim, pad=pad)
-        bounds = Box(field.bounds.lower - field.dx, field.bounds.upper + field.dx) if extrapolation == math.extrapolation.NONE else field.bounds
-        return CenteredGrid(values, bounds=bounds, extrapolation=extrapolation)
+        # ToDo if extrapolation == math.extrapolation.NONE, extend size by 1
+        # pad = 1 if extrapolation == math.extrapolation.NONE else 0
+        # bounds = Box(field.bounds.lower - field.dx, field.bounds.upper + field.dx) if extrapolation == math.extrapolation.NONE else field.bounds
+        std_widths = (0, 0)
+        if gradient_extrapolation == math.extrapolation.NONE:
+            base_widths = (abs(min(needed_shifts))+1, max(needed_shifts)+1)
+            std_widths = (1, 1)
+        padded_components = [pad(field, {dim_: base_widths if dim_ == dim else std_widths for dim_ in spatial_dims}) for dim in spatial_dims]
     elif type == StaggeredGrid:
-        assert stack_dim.name == 'vector'
-        return stagger(field, lambda lower, upper: (upper - lower) / field.dx, extrapolation)
-    raise NotImplementedError(f"{type(field)} not supported. Only CenteredGrid and StaggeredGrid allowed.")
+        assert spatial_dims == field.shape.spatial.names, f"spatial_gradient with type=StaggeredGrid requires dims=spatial, i.e. dims='{','.join(field.shape.spatial.names)}'"
+        base_widths = (base_widths[0], base_widths[1]-1)
+        padded_components = pad_for_staggered_output(field, gradient_extrapolation, field.shape.spatial.names, base_widths)
+    else:
+        raise ValueError(type)
+    shifted_components = [shift(padded_component, needed_shifts, stack_dim=None, pad=False, dims=dim) for padded_component, dim in zip(padded_components, spatial_dims)]
+    result_components = [sum([value * shift_ for value, shift_ in zip(values, shifted_component)]) / field.dx.vector[dim] for shifted_component, dim in zip(shifted_components, field.shape.spatial.names)]
+    if type == CenteredGrid:
+        result = stack(result_components, stack_dim)
+    else:
+        assert stack_dim.name == 'vector', f"spatial_gradient with type=StaggeredGrid requires stack_dim.name == 'vector' but got '{stack_dim.name}'"
+        result = StaggeredGrid(math.stack([component.values for component in result_components], channel(vector=spatial_dims)), bounds=field.bounds, extrapolation=gradient_extrapolation)
+    result = result.with_extrapolation(gradient_extrapolation)
+    if implicit:
+        implicit.x0 = result
+        result = solve_linear(_lhs_for_implicit_scheme, result, solve=implicit, values_rhs=values_rhs, needed_shifts_rhs=needed_shifts_rhs, stack_dim=stack_dim, staggered_output=type != CenteredGrid)
+    if type == CenteredGrid and gradient_extrapolation == math.extrapolation.NONE:
+        result = result.with_bounds(Box(field.bounds.lower - field.dx, field.bounds.upper + field.dx))
+    else:
+        result = result.with_bounds(field.bounds)
+    return result
 
 
-def shift(grid: CenteredGrid, offsets: tuple, stack_dim: Shape = channel('shift')):
+def _ex_map_f(ext_dict: dict):
+    def f(ext: Extrapolation):
+        return ext_dict[ext.__repr__()] if ext.__repr__() in ext_dict else ext
+    return f
+
+
+# @jit_compile_linear(auxiliary_args="values_rhs, needed_shifts_rhs, stack_dim, staggered_output")
+@jit_compile(auxiliary_args="values_rhs, needed_shifts_rhs, stack_dim, staggered_output")  # ToDo the matrix generation gives incorrect results in 2.3.0
+def _lhs_for_implicit_scheme(x, values_rhs, needed_shifts_rhs, stack_dim, staggered_output=False):
+    result = []
+    for dim, component in zip(x.shape.only(math.spatial).names, unstack(x, stack_dim.name)):
+        shifted = shift(component, needed_shifts_rhs, stack_dim=None, dims=dim)
+        result.append(sum([value * shift_ for value, shift_ in zip(values_rhs, shifted)]))
+
+    if staggered_output:
+        result = x.with_values(math.stack([component.values for component in result], channel('vector')))
+    else:
+        result = stack(result, stack_dim)
+
+    return result
+
+
+def pad_for_staggered_output(field: CenteredGrid, output_extrapolation: Extrapolation, dims: tuple, base_widths: tuple):
+    padded_components = []
+    for dim in dims:
+        border_valid = output_extrapolation.valid_outer_faces(dim)
+        padding_widths = (border_valid[0] + base_widths[0], border_valid[1] + base_widths[1])
+        padded_components.append(pad(field, {dim: padding_widths}))
+
+    return padded_components
+
+
+def shift(grid: CenteredGrid, offsets: tuple, stack_dim: Optional[Shape] = channel('shift'), dims=spatial, pad=True):
     """
     Wraps :func:`math.shift` for CenteredGrid.
 
@@ -91,8 +245,18 @@ def shift(grid: CenteredGrid, offsets: tuple, stack_dim: Shape = channel('shift'
       offsets: tuple: 
       stack_dim:  (Default value = 'shift')
     """
-    data = math.shift(grid.values, offsets, padding=grid.extrapolation, stack_dim=stack_dim)
-    return [CenteredGrid(data[i], bounds=grid.bounds, extrapolation=grid.extrapolation) for i in range(len(offsets))]
+    if pad:
+        padding = grid.extrapolation
+        new_bounds = grid.bounds
+    else:
+        padding = None
+        max_lower_shift = min(offsets) if min(offsets) < 0 else 0
+        max_upper_shift = max(offsets) if max(offsets) > 0 else 0
+        w_lower = math.wrap([max_lower_shift if dim in dims else 0 for dim in grid.shape.spatial.names])
+        w_upper = math.wrap([max_upper_shift if dim in dims else 0 for dim in grid.shape.spatial.names])
+        new_bounds = Box(grid.box.lower - w_lower * grid.dx, grid.box.upper - w_upper * grid.dx)
+    data = math.shift(grid.values, offsets, dims=dims, padding=padding, stack_dim=stack_dim)
+    return [type(grid)(data[i], bounds=new_bounds, extrapolation=grid.extrapolation) for i in range(len(offsets))]
 
 
 def stagger(field: CenteredGrid,
@@ -151,7 +315,7 @@ def stagger(field: CenteredGrid,
         raise ValueError(type)
 
 
-def divergence(field: Grid) -> CenteredGrid:
+def divergence(field: Grid, order=2, implicit: Solve = None) -> CenteredGrid:
     """
     Computes the divergence of a grid using finite differences.
 
@@ -162,28 +326,69 @@ def divergence(field: Grid) -> CenteredGrid:
 
     Args:
         field: vector field as `CenteredGrid` or `StaggeredGrid`
+        order: Spatial order of accuracy.
+            Higher orders entail larger stencils and more computation time but result in more accurate results assuming a large enough resolution.
+            Supported: 2 explicit, 4 explicit, 6 implicit.
+        implicit: When a `Solve` object is passed, performs an implicit operation with the specified solver and tolerances.
+            Otherwise, an explicit stencil is used.
 
     Returns:
         Divergence field as `CenteredGrid`
     """
-    if isinstance(field, StaggeredGrid):
-        field = bake_extrapolation(field)
-        components = []
-        for dim in field.shape.spatial.names:
-            div_dim = math.spatial_gradient(field.values.vector[dim], field.dx, 'forward', None, dims=dim, stack_dim=None)
-            components.append(div_dim)
-        data = math.sum(components, dim='0')
-        return CenteredGrid(data, bounds=field.bounds, extrapolation=field.extrapolation.spatial_gradient())
-    elif isinstance(field, CenteredGrid):
-        left, right = shift(field, (-1, 1), stack_dim=batch('div_'))
-        grad = (right - left) / (field.dx * 2)
-        components = [grad.vector[i].div_[i] for i in range(grad.div_.size)]
-        result = sum(components)
-        if field.extrapolation == math.extrapolation.NONE:
-            result = result.with_bounds(Box(field.bounds.lower + field.dx, field.bounds.upper - field.dx))
-        return result
+
+    extrap_map = {}
+    if not implicit:
+        if order == 2:
+            if isinstance(field, CenteredGrid):
+                values, needed_shifts = [-1 / 2, 1 / 2], (-1, 1)
+            else:
+                values, needed_shifts = [-1, 1], (0, 1)
+
+        elif order == 4:
+            if isinstance(field, CenteredGrid):
+                values, needed_shifts = [1 / 12, -2 / 3, 2 / 3, -1 / 12], (-2, -1, 1, 2)
+            else:
+                values, needed_shifts = [1 / 24, -27 / 24, 27 / 24, -1 / 24], (-1, 0, 1, 2)
     else:
-        raise NotImplementedError(f"{type(field)} not supported. Only StaggeredGrid allowed.")
+        extrap_map_rhs = {}
+        if order == 6:
+            extrap_map['symmetric'] = combine_by_direction(REFLECT, SYMMETRIC)
+            extrap_map_rhs['symmetric'] = combine_by_direction(ANTIREFLECT, ANTISYMMETRIC)
+
+            if isinstance(field, CenteredGrid):
+                values, needed_shifts = [-1 / 36, -14 / 18, 14 / 18, 1 / 36], (-2, -1, 1, 2)
+                values_rhs, needed_shifts_rhs = [1 / 3, 1, 1 / 3], (-1, 0, 1)
+
+            else:
+                values, needed_shifts = [-17 / 186, -63 / 62, 63 / 62, 17 / 186], (-1, 0, 1, 2)
+                values_rhs, needed_shifts_rhs = [9 / 62, 1, 9 / 62], (-1, 0, 1)
+    base_widths = (abs(min(needed_shifts)), max(needed_shifts))
+    field.with_extrapolation(extrapolation.map(_ex_map_f(extrap_map), field.extrapolation))  # ToDo does this line do anything?
+    spatial_dims = field.shape.spatial.names
+    if isinstance(field, StaggeredGrid):
+        base_widths = (base_widths[0]+1, base_widths[1])
+        padded_components = []
+        for dim, component in zip(field.shape.spatial.names, unstack(field, 'vector')):
+            border_valid = field.extrapolation.valid_outer_faces(dim)
+            padding_widths = (base_widths[0] - border_valid[0], base_widths[1] - border_valid[1])
+            padded_components.append(pad(component, {dim: padding_widths}))
+    elif isinstance(field, CenteredGrid):
+        padded_components = [pad(component, {dim: base_widths}) for dim, component in zip(spatial_dims, unstack(field, 'vector'))]
+        if field.extrapolation == math.extrapolation.NONE:
+            padded_components = [pad(component, {dim_: (0, 0) if dim_ == dim else (-1, -1) for dim_ in spatial_dims}) for dim, component in zip(spatial_dims, padded_components)]
+    shifted_components = [shift(padded_component, needed_shifts, None, pad=False, dims=dim) for padded_component, dim in zip(padded_components, spatial_dims)]
+    result_components = [sum([value * shift for value, shift in zip(values, shifted_component)]) / field.dx.vector[dim] for shifted_component, dim in zip(shifted_components, spatial_dims)]
+    if implicit:
+        result_components = stack(result_components, channel('vector'))
+        result_components.with_values(result_components.values._cache())
+        implicit.x0 = field
+        result_components = solve_linear(_lhs_for_implicit_scheme, result_components, solve=implicit, values_rhs=values_rhs, needed_shifts_rhs=needed_shifts_rhs, stack_dim=channel('vector'))
+        result_components = unstack(result_components, 'vector')
+    result_components = [component.with_bounds(field.bounds) for component in result_components]
+    result = sum(result_components)
+    if field.extrapolation == math.extrapolation.NONE and isinstance(field, CenteredGrid):
+        result = result.with_bounds(Box(field.bounds.lower + field.dx, field.bounds.upper - field.dx))
+    return result
 
 
 def curl(field: Grid, type: type = CenteredGrid):
@@ -319,7 +524,7 @@ def pad(grid: GridType, widths: int or tuple or list or dict) -> GridType:
         widths = {axis: (width if isinstance(width, (tuple, list)) else (width, width)) for axis, width in zip(grid.shape.spatial.names, widths)}
     else:
         assert isinstance(widths, dict)
-    widths_list = [widths[axis] for axis in grid.shape.spatial.names]
+    widths_list = [widths[axis] if axis in widths.keys() else (0, 0) for axis in grid.shape.spatial.names]
     if isinstance(grid, Grid):
         data = math.pad(grid.values, widths, grid.extrapolation, bounds=grid.bounds)
         w_lower = math.wrap([w[0] for w in widths_list])
@@ -402,10 +607,9 @@ def concat(fields: List[SampledFieldType] or Tuple[SampledFieldType, ...], dim:
         values = math.concat([f.values for f in fields], dim)
         return fields[0].with_values(values)
     elif isinstance(fields[0], PointCloud):
-        elements = geom.concat([f.elements for f in fields], dim, sizes=[f.shape.get_size(dim) for f in fields])
+        elements = geom.concat([f.elements for f in fields], dim)
         values = math.concat([math.expand(f.values, f.shape.only(dim)) for f in fields], dim)
-        colors = math.concat([math.expand(f.color, f.shape.only(dim)) for f in fields], dim)
-        return PointCloud(elements=elements, values=values, color=colors, extrapolation=fields[0].extrapolation, add_overlapping=fields[0]._add_overlapping, bounds=fields[0]._bounds)
+        return PointCloud(elements=elements, values=values, extrapolation=fields[0].extrapolation, add_overlapping=fields[0]._add_overlapping, bounds=fields[0]._bounds)
     raise NotImplementedError(type(fields[0]))
 
 
@@ -419,6 +623,7 @@ def stack(fields, dim: Shape, dim_bounds: Box = None):
     Args:
         fields: List of matching `SampledField` instances.
         dim: Stack dimension as `Shape`. Size is ignored.
+        dim_bounds: `Box` defining the physical size for `dim`.
 
     Returns:
         `SampledField` matching stacked fields.
@@ -436,10 +641,9 @@ def stack(fields, dim: Shape, dim_bounds: Box = None):
         else:
             return fields[0].with_values(values)
     elif isinstance(fields[0], PointCloud):
-        elements = geom.stack([f.elements for f in fields], dim=dim)
-        values = math.stack([f.values for f in fields], dim=dim)
-        colors = math.stack([f.color for f in fields], dim=dim)
-        return PointCloud(elements=elements, values=values, color=colors, extrapolation=fields[0].extrapolation, add_overlapping=fields[0]._add_overlapping, bounds=fields[0]._bounds)
+        elements = geom.stack([f.elements for f in fields], dim)
+        values = math.stack([f.values for f in fields], dim)
+        return PointCloud(elements=elements, values=values, extrapolation=fields[0].extrapolation, add_overlapping=fields[0]._add_overlapping, bounds=fields[0]._bounds)
     raise NotImplementedError(type(fields[0]))
 
 
@@ -519,6 +723,7 @@ def _auto_resample(*fields: Field):
 
 def vec_length(field: SampledField):
     """ See `phi.math.vec_abs()` """
+    assert isinstance(field, SampledField), f"SampledField required but got {type(field).__name__}"
     if isinstance(field, StaggeredGrid):
         field = field.at_centers()
     return field.with_values(math.vec_abs(field.values))
@@ -567,7 +772,7 @@ def discretize(grid: Grid, filled_fraction=0.25):
     return grid.with_values(filled_t)
 
 
-def integrate(field: Field, region: Geometry, scheme: Scheme = Scheme()) -> Tensor:
+def integrate(field: Field, region: Geometry, **kwargs) -> Tensor:
     """
     Computes *∫<sub>R</sub> f(x) dx<sup>d</sup>* , where *f* denotes the `Field`, *R* the `region` and *d* the number of spatial dimensions (`d=field.shape.spatial_rank`).
     Depending on the `sample` implementation for `field`, the integral may be a rough approximation.
@@ -577,37 +782,14 @@ def integrate(field: Field, region: Geometry, scheme: Scheme = Scheme()) -> Tens
     Args:
         field: `Field` to integrate.
         region: Region to integrate over.
-        scheme: Numerical scheme.
+        **kwargs: Specify numerical scheme.
 
     Returns:
         Integral as `phi.Tensor`
     """
     if not isinstance(field, CenteredGrid):
         raise NotImplementedError()
-    return field._sample(region, scheme=scheme) * region.volume
-
-
-def tensor_as_field(t: Tensor):
-    """
-    Interpret a `Tensor` as a `CenteredGrid` or `PointCloud` depending on its dimensions.
-
-    Unlike the `CenteredGrid` constructor, this function will have the values sampled at integer points for each spatial dimension.
-
-    Args:
-        t: `Tensor` with either `spatial` or `instance` dimensions.
-
-    Returns:
-        `CenteredGrid` or `PointCloud`
-    """
-    if instance(t):
-        bounds = data_bounds(t)
-        return PointCloud(t, bounds=Cuboid(bounds.center, bounds.half_size * 1.2).box())
-    elif spatial(t):
-        return CenteredGrid(t, 0, bounds=Box(math.const_vec(-0.5, spatial(t)), wrap(spatial(t), channel('vector')) - 0.5))
-    elif 'vector' in t.shape:
-        return PointCloud(math.expand(t, instance(points=1)), bounds=Cuboid(t, half_size=math.const_vec(1, t.shape['vector'])).box())
-    else:
-        raise ValueError(f"Cannot create field from tensor with shape {t.shape}. Requires at least one spatial, instance or vector dimension.")
+    return field._sample(region, **kwargs) * region.volume
 
 
 def pack_dims(field: SampledFieldType,
@@ -646,3 +828,23 @@ def support(field: SampledField, list_dim: Shape or str = instance('nonzero')) -
         `Tensor` with shape `(list_dim, vector)`
     """
     return field.points[math.nonzero(field.values, list_dim=list_dim)]
+
+
+def mask(obj: SampledFieldType or Geometry) -> SampledFieldType:
+    """
+    Returns a `Field` that masks the inside (or non-zero values when `obj` is a grid) of a physical object.
+    The mask takes the value 1 inside the object and 0 outside.
+    For `CenteredGrid` and `StaggeredGrid`, the mask labels non-zero non-NaN entries as 1 and all other values as 0
+
+    Returns:
+        `Grid` type or `PointCloud`
+    """
+    if isinstance(obj, PointCloud):
+        return PointCloud(obj.elements, 1, math.extrapolation.remove_constant_offset(obj.extrapolation), bounds=obj.bounds)
+    elif isinstance(obj, Geometry):
+        return PointCloud(obj, 1, 0)
+    elif isinstance(obj, CenteredGrid):
+        values = math.cast(obj.values != 0, int)
+        return obj.with_values(values)
+    else:
+        raise ValueError(obj)
