@@ -20,8 +20,7 @@ from .physics import fluid, advect, diffuse
 # Classes
 from .math import Tensor, DType, Solve
 from .geom import Geometry, Sphere, Box, Cuboid
-from .field import Grid, CenteredGrid, StaggeredGrid, GeometryMask, SoftGeometryMask, HardGeometryMask, Noise, PointCloud, Scene
-from .field.numerical import Scheme
+from .field import Grid, CenteredGrid, StaggeredGrid, mask, Noise, PointCloud, Scene, resample, GeometryMask, SoftGeometryMask, HardGeometryMask
 from .vis import Viewer
 from .physics.fluid import Obstacle
 
@@ -31,9 +30,10 @@ from .math import PI, INF, NAN
 # Functions
 from .math import (
     wrap, tensor, vec,  # Tensor creation
-    shape, spatial, channel, batch, instance, non_spatial, non_channel, non_batch, non_instance,  # Shape functions (magic)
+    shape, spatial, channel, batch, instance, dual,
+    non_spatial, non_channel, non_batch, non_instance, non_dual,  # Shape functions (magic)
     unstack, stack, concat, expand, rename_dims, pack_dims, unpack_dim, flatten, cast,  # Magic Ops
-    jit_compile, jit_compile_linear, minimize, functional_gradient, solve_linear, solve_nonlinear, iterate,  # jacobian, hessian, custom_gradient # Functional magic
+    jit_compile, jit_compile_linear, minimize, functional_gradient, solve_linear, solve_nonlinear, iterate, identity,  # jacobian, hessian, custom_gradient # Functional magic
 )
 from .geom import union
 from .vis import show, view, control, plot
